/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.army.hierarchy;

import java.awt.Image;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import fb.army.entities.Entity;
import fb.army.entities.Leader;
import fb.army.entities.SummonedCreature;
import fb.army.entities.Unit;
import fb.main.FB;
import fb.network.Client;
import fb.network.MessagesConstants;
import fb.util.D100;
import fb.util.Resources;
import fb.util.StringTools;

/**
 * an army command
 * 
 * @author farfpeuf
 * 
 */
public class Command extends HierarchyNode {
	/**
	 * icon used in tree table
	 */
	protected static Icon minicon;

	/**
	 * command name
	 */
	protected String name;

	/**
	 * army to which this command belongs
	 */
	protected Army army;

	/**
	 * command leader
	 */
	protected Leader leader;

	/**
	 * lists
	 */
	protected Vector individuals, allEntities;

	protected Vector activeEntities, units;

	/**
	 * current orders
	 */
	protected char order = 'o';

	/**
	 * has command changed its orders this turn ?
	 */
	protected boolean changedOrders = false;

	/**
	 * constructor initialize lists
	 */
	public Command() {
		individuals = new Vector();
		units = new Vector();
		activeEntities = new Vector();
		allEntities = new Vector();
	}

	/**
	 * build command with another command
	 * 
	 * @param arm
	 *            army to which this command belongs
	 * @param other
	 *            the other command
	 */
	public Command(Army arm, Command other) {
		this();
		this.army = arm;
		name = Resources.getString("Othercommand");
		List entities = other.getActiveEntities();
		Entity entity;
		for (int i = 0; i < entities.size(); i++) {
			entity = (Entity) entities.get(i);
			addEntity(entity);
		}
	}

	/**
	 * reset command after end of battle
	 * 
	 * @param winner
	 *            did the command army won the battle
	 */
	public void endBattle(boolean winner) {
		List oldEntities = new ArrayList(allEntities);
		allEntities.clear();
		individuals.clear();
		units.clear();
		activeEntities.clear();

		Entity entity;
		for (int i = 0; i < oldEntities.size(); i++) {
			entity = (Entity) oldEntities.get(i);
			if (entity.getState() != 'a'
					&& !(entity instanceof SummonedCreature)) {
				entity.endBattle(winner);
				if (entity.getNumberOfWarriors() > 0) {
					addEntity(entity);
				}
			}
		}
	}

	/**
	 * calculate cost
	 */
	public int getCost() {
		int cost = 0;
		for (int i = 0; i < activeEntities.size(); i++) {
			cost += ((Entity) activeEntities.get(i)).getCost();
		}
		return cost;
	}

	/**
	 * number of warriors in this command
	 */
	public int getNumberOfWarriors() {
		int number = 0;
		for (int i = 0; i < activeEntities.size(); i++) {
			number += ((Entity) activeEntities.get(i)).getNumberOfWarriors();
		}
		return number;
	}

	/**
	 * @return maximum movement
	 */
	public int movement() {
		int movement = Integer.MAX_VALUE;
		for (int i = 0; i < activeEntities.size(); i++) {
			movement = Math.min(movement, ((Entity) activeEntities.get(i))
					.getMovement());
		}
		return movement;
	}

	/**
	 * returns command as a string to save it to a file
	 * 
	 * @return command as a string
	 */

	public String saveString() {
		String savestring = "g" + '\t' + name + '\t' + order + '\t'
				+ (changedOrders ? 'y' : 'n') + '\n';
		for (int i = 0; i < allEntities.size(); i++) {
			savestring += (((Entity) allEntities.get(i)).saveString() + '\n');
		} // for
		return savestring;
	}

	/**
	 * load command from a StringBuffer
	 * 
	 * @param buffy
	 *            the StringBuffer
	 */
	public void loadString(StringBuffer buffy) {
		name = StringTools.readBuffer(buffy);
		order = StringTools.readBufferChar(buffy);
		changedOrders = (StringTools.readBufferChar(buffy) == 'y') ? true
				: false;
	}

	/**
	 * recruit a new entity
	 * 
	 * @param e
	 *            the new entity
	 */
	public void recruitEntity(Entity e) {
		activeEntities.add(e);
		allEntities.add(e);
		army.addEntity(e);
		e.setCommand(this);
	}

	/**
	 * fire an entity
	 * 
	 * @param e
	 *            the fired entity
	 */
	public void fireEntity(Entity e) {
		activeEntities.remove(e);
		allEntities.remove(e);
		if (army != null) {
			army.removeEntity(e);
		}
		e.setCommand(null);
	}

	/**
	 * add an active entity
	 * 
	 * @param e
	 *            the active entity
	 */
	public void addActiveEntity(Entity e) {
		activeEntities.add(e);

		if (e.isIndividual()) {
			individuals.add(e);
		} else {
			units.add(e);
		}
		army.addEntity(e);
	}

	/**
	 * add an entity
	 * 
	 * @param e
	 *            the entity to add
	 */
	public void addEntity(Entity e) {
		allEntities.add(e);
		e.setCommand(this);
		if (e.isActive()) {
			addActiveEntity(e);
		}

	}

	/**
	 * remove an active entity
	 * 
	 * @param e
	 *            the active entity
	 */
	public void removeActiveEntity(Entity e) {
		activeEntities.remove(e);
		if (army != null) {
			army.removeEntity(e);
		}
		if (e.isIndividual()) {
			individuals.remove(e);
		} else {
			units.remove(e);
		}

		if (activeEntities.size() == 0) {
			if (army != null) {
				army.removeCommand(this);
			}
		} else if (leader != null && activeEntities.size() == 1
				&& leader.isActive()) {
			leader.rout();
		}
	}

	/**
	 * test command morale
	 * 
	 */
	public void moraleTest() {
		int score = D100.roll();

		if (leader != null && leader.getState() == 'n') {
			score += (leader.getLeadership() * 15);
		}

		if (army.getWarchief() == null || army.getWarchief().getState() == 'r') {
			score -= 40;
		}

		if (army.getOmens() == 'b') {
			score -= 15;
		}
		if (suffersBadLight()) {
			score -= 15;
		}
		if (army.getOmens() == 'g') {
			score += 15;
		}
		score -= (army.routedAnnihilatedCommands() * 15);
		groupReaction(score);
	}

	/**
	 * caculate reaction to morale test
	 * 
	 * @param score
	 *            morale test score
	 */
	public void groupReaction(int score) {

		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.GROUPREACTION + army.getGmId() + "#"
							+ getArmIndex() + "#" + score);
		} else {
			doGroupReaction(score);
		}
	}

	/**
	 * get index in army commands list
	 * 
	 * @return
	 */
	public int getArmIndex() {

		return army.getCommands().indexOf(this);
	}

	/**
	 * do caculate reaction to morale test
	 * 
	 * @param score
	 *            morale test score
	 */
	public void doGroupReaction(int score) {

		// FB.debug("doGroupReaction " + getFullName() + " " + score);

		if (score >= 95) {
		} else if (score >= 65) {
			setDisorganized();
		} else if (score >= 50) {
			setShaken();
		} else {
			rout();
		}

	}

	/**
	 * rout the command
	 * 
	 */
	public void rout() {
		String message = Resources.getString("grouprout");
		message = StringTools.replaceAll(message, "<group name>", name);
		FB.getInstance().message(message, army.getColor());
		for (int i = 0; i < activeEntities.size(); i++) {
			((Entity) activeEntities.get(i)).rout();
		}
	}

	/**
	 * set all units shaken
	 * 
	 */
	public void setShaken() {
		for (int i = 0; i < units.size(); i++) {
			((Unit) units.get(i)).doSetMorale(3);
		}
	}

	/**
	 * set all units disorganized
	 * 
	 */
	public void setDisorganized() {
		for (int i = 0; i < units.size(); i++) {
			((Unit) units.get(i)).disorganize();
		}
	}

	/**
	 * set all units in bloodlust
	 * 
	 */
	public void setBloodlust() {
		for (int i = 0; i < units.size(); i++) {
			((Unit) units.get(i)).doSetMorale(10);
		}
	}

	/**
	 * @return number of units in bloodlust
	 */
	public int nbBloodlustUnits() {
		int result = 0;
		for (int i = 0; i < units.size(); i++) {
			if (((Unit) units.get(i)).getState() == 'b') {
				result++;
			}
		}
		return result;
	}

	/**
	 * @return number of shaken units
	 */
	public int nbShakenUnits() {
		int result = 0;
		for (int i = 0; i < units.size(); i++) {
			if (((Unit) units.get(i)).getState() == 's') {
				result++;
			}
		}
		return result;
	}

	/**
	 * @return true if command has disorganized units
	 */
	public boolean hasDisorganizedUnits() {
		Unit unit;
		for (int i = 0; i < units.size(); i++) {
			unit = (Unit) units.get(i);
			if (unit.isDisorganized() && unit.isActive()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * @return true if one of the units is fanatic
	 */
	public boolean isFanatic() {
		Unit unit;
		for (int i = 0; i < units.size(); i++) {
			unit = (Unit) units.get(i);
			if (unit.getType() == 'f' && unit.isActive()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * @return true if one of the entities suffers from bad light
	 */
	public boolean suffersBadLight() {
		Entity entity;
		for (int i = 0; i < activeEntities.size(); i++) {
			entity = (Entity) activeEntities.get(i);
			if (entity.suffersBadLight() && entity.isActive()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * change command orders
	 * 
	 * @param neworder
	 *            new orders
	 */
	public void changeOrder(char neworder) {
		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.NEWORDER + army.getGmId() + "#" + getArmIndex()
							+ "#" + Character.toString(neworder));
		} else {
			doChangeOrder(neworder);
		}
	}

	/**
	 * do change command orders
	 * 
	 * @param neworder
	 *            new orders
	 */
	public void doChangeOrder(char neworder) {

		setOrder(neworder);
		changedOrders = true;
		String orderstring;

		if (order == 'a') {
			orderstring = "attack";
		} else if (order == 'o') {
			orderstring = "oppose";
		} else {
			orderstring = "hold";
		}
		for (int i = 0; i < units.size(); i++) {
			((Unit) (units.get(i))).doChangeOrder(order);
		}
		String message = Resources.getString("groupneworder");
		message = StringTools.replaceAll(StringTools.replaceAll(message,
				"<group name>", name), "<order name>", Resources
				.getString(orderstring));
		FB.getInstance().message(message, army.getColor());

	}

	/**
	 * Sets the order.
	 * 
	 * @param order
	 *            The order to set
	 */
	public void setOrder(char order) {
		char oldValue = this.order;
		this.order = order;
		firePropertyChange("order", oldValue, order);
	}

	/**
	 * @see javax.swing.tree.TreeNode#children()
	 */
	public Enumeration children() {
		return activeEntities.elements();
	}

	/**
	 * @see javax.swing.tree.TreeNode#getAllowsChildren()
	 */
	public boolean getAllowsChildren() {
		return true;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getChildAt(int)
	 */
	public TreeNode getChildAt(int childIndex) {
		return (TreeNode) activeEntities.get(childIndex);
	}

	/**
	 * @see javax.swing.tree.TreeNode#getChildCount()
	 */
	public int getChildCount() {
		return activeEntities.size();
	}

	/**
	 * @see javax.swing.tree.TreeNode#getIndex(TreeNode)
	 */
	public int getIndex(TreeNode node) {
		return activeEntities.indexOf(node);
	}

	/**
	 * @see javax.swing.tree.TreeNode#getParent()
	 */
	public TreeNode getParent() {
		return army;
	}

	/**
	 * @see javax.swing.tree.TreeNode#isLeaf()
	 */
	public boolean isLeaf() {
		return false;
	}

	/**
	 * fire this command
	 * 
	 */
	public void fire() {
		army.fireCommand(this);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#insert(MutableTreeNode, int)
	 */
	public void insert(MutableTreeNode child, int index) {
		recruitEntity((Entity) child);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#remove(int)
	 */
	public void remove(int index) {
		fireEntity((Entity) activeEntities.get(index));
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#remove(MutableTreeNode)
	 */
	public void remove(MutableTreeNode node) {
		fireEntity((Entity) node);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#removeFromParent()
	 */
	public void removeFromParent() {
		army.remove(this);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#setParent(MutableTreeNode)
	 */
	public void setParent(MutableTreeNode newParent) {
		army = (Army) newParent;
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#setUserObject(Object)
	 */
	public void setUserObject(Object object) {
	}

	/**
	 * @return command full name
	 */
	public String getFullName() {
		return name;
	}

	/**
	 * @return icon used for tree table
	 */
	public Icon getMinicon() {
		if (null == minicon) {
			minicon = new ImageIcon((Resources.getSign("group")).getImage()
					.getScaledInstance(20, 20, Image.SCALE_SMOOTH));

		}
		return minicon;
	}

	/**
	 * Returns the individuals.
	 * 
	 * @return Vector
	 */
	public Vector getIndividuals() {
		return individuals;
	}

	/**
	 * Sets the individuals.
	 * 
	 * @param individuals
	 *            The individuals to set
	 */
	public void setIndividuals(Vector individuals) {
		this.individuals = individuals;
	}

	/**
	 * @return true if this command is routed or annihilated
	 */
	public boolean isRoutedAnnihilated() {
		return activeEntities.size() == 0;
	}

	/**
	 * @return true if this command is annihilated
	 */
	public boolean isAnnihilated() {
		Entity entity;
		for (int i = 0; i < allEntities.size(); i++) {
			entity = (Entity) allEntities.get(i);

			if (entity.getState() != 'a') {
				return false;
			}
		}
		return true;
	}

	/**
	 * @return number of models in command
	 */
	public int numberOfModels() {
		int number = 0;
		Iterator iter = allEntities.iterator();
		while (iter.hasNext()) {
			number += ((Entity) iter.next()).numberOfModels();
		}
		return number;
	}

	/**
	 * return number of entities of specified class
	 * 
	 * @param cl
	 *            class to search for
	 * @return number of entities of specified class
	 */
	public int numberOf(Class cl) {
		int number = 0;
		Iterator iter = allEntities.iterator();
		while (iter.hasNext()) {
			if (cl.isInstance(iter.next())) {
				number++;
			}
		}
		return number;
	}

	/**
	 * check command content before saving
	 * 
	 * @return null if command is ok or an error message
	 */
	public String check() {
		String message = "";
		int numberOfLeader = numberOf((new Leader(null, 0)).getClass());
		if (numberOfLeader == 0) {
			message += Resources.getString("groupNoLeader");
		} else if (numberOfLeader > 1) {
			message += Resources.getString("groupMoreThanOneLeader");
		}

		List racesList = races();
		if (racesList.contains("goblin")
				&& (racesList.contains("woodelf") || racesList
						.contains("high-elf"))) {
			message += Resources.getString("noGoblinAndElf");
		}
		if (racesList.contains("dwarf") && racesList.contains("orc")) {
			message += Resources.getString("noDwarfAndOrc");
		}

		message = StringTools.replaceAll(message, "<group name>", name);

		return message;
	}

	/**
	 * @return list of all races in this command
	 */
	public List races() {
		List racesList = new ArrayList();
		Iterator iter = allEntities.iterator();
		String race;
		while (iter.hasNext()) {
			race = ((Entity) iter.next()).getRace();
			if (!racesList.contains(race)) {
				racesList.add(race);
			}
		}
		return racesList;
	}

	/**
	 * @return Returns the allEntities.
	 */
	public synchronized Vector getAllEntities() {
		return allEntities;
	}

	/**
	 * @return Returns the activeEntities.
	 */
	public Vector getActiveEntities() {
		return activeEntities;
	}

	/**
	 * @return Returns the army.
	 */
	public Army getArmy() {
		return army;
	}

	/**
	 * @return Returns the leader.
	 */
	public Leader getLeader() {
		return leader;
	}

	/**
	 * @return Returns the order.
	 */
	public char getOrder() {
		return order;
	}

	/**
	 * @return Returns the units.
	 */
	public Vector getUnits() {
		return units;
	}

	/**
	 * @param army
	 *            The army to set.
	 */
	public void setArmy(Army army) {
		this.army = army;
	}

	/**
	 * @param leader
	 *            The leader to set.
	 */
	public void setLeader(Leader leader) {
		this.leader = leader;
	}

	/**
	 * @return true if command changed its orders this turn
	 */
	public boolean hasChangedOrders() {
		return changedOrders;
	}

	/**
	 * reset changedOrders value
	 * 
	 */
	public void resetChangedOrders() {
		changedOrders = false;
	}

	/**
	 * @param name
	 *            The name to set.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return list of races that can be added to this command
	 */
	public String[] getAvailableRaces() {
		String[] catalog = new String[] { Resources.getString("ae"
				+ army.getRace()) };
		return catalog;
	}

}
