/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.artificialintelligence;

import java.awt.geom.Area;
import java.util.Iterator;
import java.util.List;

import fb.army.entities.Entity;

/**
 * AI thread that chooses the best target among a set of availables targets
 * 
 * @author farfpeuf
 * 
 */
public class TargetFinder extends WeightChooser {

	/**
	 * the entity for which a target is searched
	 */
	protected Entity entity;

	/**
	 * optimization avoiding to calculate n-times the obstacles between a actor and potential target 
	 */
	private Area obstacles = null;

	/**
	 * optimization avoiding to calculate n-times the closets target
	 */
	private Entity closestTarget = null;
	
	/**
	 * create a target finder the best target among a set of availables targets
	 * 
	 * @param targetables
	 *            availables targets
	 * @param nbD6
	 *            numbers of d6 to roll to get the random factor
	 * @param actor
	 *            the entity for which a target is searched
	 */
	public TargetFinder(List targetables, int nbD6, Entity actor) {
		this(targetables, nbD6, nbD6 * 6, actor);
	}

	/**
	 * create a target finder to find the best target among a set of availables targets
	 * 
	 * @param targetables
	 *            availables targets
	 * @param nbD6
	 *            numbers of d6 to roll to get the random factor
	 * @param minWeight
	 *            the minimum weight to have to be a good target (if none of the
	 *            availables targets obtain more than this weight, no target
	 *            will be selected)
	 * @param actor
	 *            the entity for which a target is searched
	 */
	public TargetFinder(List targetables, int nbD6, int minWeight, Entity actor) {
		super(targetables, nbD6, minWeight);
		setName("TargetFinder");
		this.entity = actor;
	}

	/**
	 * get the best target found by the algorythm
	 * 
	 * @return the best target found
	 */
	public Entity getTarget() {
		return (Entity) getChosenElement();
	}

	/**
	 * find the closest target from the entity for which a target is searched
	 * @return the closest target among the set of availables targets
	 */
	protected Entity getClosestTarget() {
		if (null == closestTarget) {
			Entity potentialTarget;
			Iterator iter = availableElements.iterator();
			int dist, mindist = 9999;
			while (iter.hasNext()) {
				potentialTarget = (Entity) iter.next();

				dist = entity.distanceTo(potentialTarget);
				if (dist < mindist) {
					mindist = dist;
					closestTarget = potentialTarget;
				}

			}
		}
		return closestTarget;
	}

	/**
	 * get the obstacles between a potential target and the entity for which a target is searched
	 * @param potentialTarget  a potential target 
	 * @return the obstacles area between the 2 entities
	 */
	protected Area getMouvementObstacles(Entity potentialTarget) {
		if (null == obstacles) {
			obstacles = entity.getMovementObstacles();
		}
		Area result = new Area(obstacles);
		result.subtract(potentialTarget.getArea());
		return result;
	}


}