/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefield;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

import fb.util.ArrayListModel;
import fb.util.FileTools;

/**
 * The data model for a game map (containing obstacles, deployment areas ...)
 * 
 * @author farfpeuf
 * 
 */
public class MapModel {

	// vertical sets having a shadow
	public static final String[] verticalSetsWithShadows = new String[] { "J",
			"K", "L", "M" };

	// used by property change listener
	public static String SIDES_PROPERTY = "side";

	// is map flipping allowed
	private static boolean flippingAllowed = false;

	// map model file path
	private String path;

	// current polygon being drawn
	private Polygon poli;

	// current points being placed
	private List currentpoints;

	// map containing list of points or polygons mapped by type
	private Hashtable allSets;

	// map containing list of points for vertical sets mapped by type
	private Hashtable verticalSets;

	// list of sides
	private ArrayListModel sides;

	// map width and length in game unit
	private int mapWidth = 576, mapLength = 384;

	// current set type
	private char currenttype = 'r';

	// singleton
	private static MapModel instance;

	// flip map from left to right ? , from up to down
	private boolean flipLR, flipUD;

	// map image file name
	private String imageName;

	/**
	 * get singleton instance
	 * 
	 * @return the singleton
	 */
	public static synchronized MapModel getInstance() {
		if (null == instance) {
			instance = new MapModel();
		}
		return instance;
	}

	/**
	 * private constructor used by getinstance method
	 * 
	 */
	private MapModel() {
		verticalSets = new Hashtable();
		allSets = new Hashtable();
		currentpoints = new ArrayList();
		sides = new ArrayListModel();

	}

	/**
	 * undo previous change
	 * 
	 */
	public void undo() {
		if (Character.isLowerCase(currenttype)) {
			List list = getSet(currenttype);
			if (list != null && !list.isEmpty()) {
				list.remove(list.size() - 1);
				if (list.isEmpty()) {
					allSets.remove(String.valueOf(currenttype));
				}
			}
		} else {
			List vset = (List) verticalSets.get(String.valueOf(currenttype));
			if (vset != null && !vset.isEmpty()) {
				vset.remove(vset.size() - 1);
				if (vset.isEmpty()) {
					verticalSets.remove(String.valueOf(currenttype));
				}
			}
		}
	}

	/**
	 * reset the model
	 * 
	 */
	public void reset() {
		sides.clear();
		verticalSets.clear();
		allSets.clear();
		currentpoints.clear();
		flipLR = false;
		flipUD = false;
	}

	/**
	 * begin a new obstacle
	 * 
	 * @param type
	 *            type of obstacle
	 */
	public void newobstacle(String type) {
		currenttype = type.charAt(1);
	}

	/**
	 * load a model from a file
	 * 
	 * @param path
	 *            the file path
	 */
	public void load(String path) {

		reset();

		this.path = path;
		String ligne;
		Reader fr = null;
		BufferedReader br;
		int x, y;
		try {
			fr = FileTools.getReader(path);
			br = new BufferedReader(fr);
			imageName = br.readLine();
			ligne = br.readLine();
			mapWidth = Integer.parseInt(ligne);
			ligne = br.readLine();
			mapLength = Integer.parseInt(ligne);
			ligne = br.readLine();
			while (ligne != null) {
				if (ligne.charAt(0) == '#') {
					close();
					newobstacle(ligne);
					ligne = br.readLine();
				}
				x = Integer.parseInt(ligne);
				ligne = br.readLine();
				y = Integer.parseInt(ligne);
				ligne = br.readLine();
				currentpoints.add(new Point(x, y));
			}
			close();
			br.close();
			fr.close();
			if (flippingAllowed) {
				flip(false, true);
			}
		} // try
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * save the model to its file
	 * 
	 */
	public void save() {
		if (path != null) {
			save(path);
		}
	}

	/**
	 * save the model to a file
	 * 
	 * @param file
	 *            the file
	 */
	public void save(String file) {
		FileTools.writeFile(file, getModelString());
	}

	/**
	 * add a point to current points list
	 * 
	 * @param p
	 *            the point to add
	 */
	public void addPoint(Point p) {
		currentpoints.add(p);
		if (currentpoints.size() >= 2 && currenttype >= 'a'
				&& currenttype <= 'f') {
			close();
		}
	}

	/**
	 * return data model as a string (to save it into a file)
	 * 
	 * @return the model as a string
	 */
	public String getModelString() {
		String result = "";
		result += imageName + "\n";
		result += mapWidth + "\n";
		result += mapLength + "\n";
		Enumeration keys = allSets.keys();
		String key;
		while (keys.hasMoreElements()) {
			key = (String) keys.nextElement();
			result += getModelString((List) allSets.get(key), key.charAt(0));
		}
		keys = verticalSets.keys();
		while (keys.hasMoreElements()) {
			key = (String) keys.nextElement();
			result += getModelString((List) verticalSets.get(key), key
					.charAt(0));
		}
		return result;
	}

	/**
	 * flip the model elements
	 * 
	 * @param lr
	 *            flip from left to right ?
	 * @param ud
	 *            flip from up to down ?
	 */
	public void flip(boolean lr, boolean ud) {

		if (lr) {
			flipLR = !flipLR;
		}
		if (ud) {
			flipUD = !flipUD;
		}

		Enumeration keys = allSets.keys();
		String key;
		while (keys.hasMoreElements()) {
			key = (String) keys.nextElement();
			flip((List) allSets.get(key), lr, ud);
		}
		keys = verticalSets.keys();
		while (keys.hasMoreElements()) {
			key = (String) keys.nextElement();
			flip((List) verticalSets.get(key), lr, ud);
		}
	}

	/**
	 * flip all the geometric elements contained in a list
	 * 
	 * @param vect
	 *            the list of elements
	 * @param lr
	 *            flip from left to right ?
	 * @param ud
	 *            flip from up to down ?
	 */
	private void flip(List vect, boolean lr, boolean ud) {

		Polygon poli;
		Rectangle rect;
		Point p;
		Object element;
		int[] polix, poliy;
		for (int i = 0; i < vect.size(); i++) {
			element = vect.get(i);
			if (element instanceof Polygon) {

				poli = (Polygon) element;
				polix = poli.xpoints;
				poliy = poli.ypoints;
				for (int j = 0; j < poli.npoints; j++) {
					if (lr) {
						polix[j] = -polix[j] + mapWidth;
					}
					if (ud) {
						poliy[j] = -poliy[j] + mapLength;
					}
				}
			}
			if (element instanceof Point) {

				p = (Point) element;
				if (lr) {
					p.x = -p.x + mapWidth;
				}
				if (ud) {
					p.y = -p.y + mapLength;
				}
			}
			if (element instanceof Rectangle) {
				rect = (Rectangle) element;
				if (lr) {
					rect.x = mapWidth - rect.x - rect.width;
				}
				if (ud) {
					rect.y = mapLength - rect.y - rect.height;
				}
			}
		}
	}

	/**
	 * return a list of geometric elements as a string (to save it into a file)
	 * 
	 * @param vect
	 *            the list of elements
	 * @param type
	 *            the type of element
	 * @return the list of element as a string
	 */
	private String getModelString(List vect, char type) {
		String result = "";
		Polygon poli;
		Rectangle rect;
		Point p;
		Object element;
		int[] polix, poliy;
		for (int i = 0; i < vect.size(); i++) {
			element = vect.get(i);
			if (element instanceof Polygon) {
				result += "#" + String.valueOf(type) + "\n";
				poli = (Polygon) element;
				polix = poli.xpoints;
				poliy = poli.ypoints;
				for (int j = 0; j < poli.npoints; j++) {
					result += Integer.toString(polix[j]) + "\n";
					result += Integer.toString(poliy[j]) + "\n";
				}
			}
			if (element instanceof Point) {
				if (i == 0) {
					result += "#" + String.valueOf(type) + "\n";
				}
				p = (Point) element;
				result += Integer.toString(p.x) + "\n";
				result += Integer.toString(p.y) + "\n";
			}
			if (element instanceof Rectangle) {
				result += "#" + String.valueOf(type) + "\n";
				rect = (Rectangle) element;
				result += Integer.toString(rect.x) + "\n";
				result += Integer.toString(rect.y) + "\n";
				result += Integer.toString(rect.x + rect.width) + "\n";
				result += Integer.toString(rect.y + rect.height) + "\n";
			}
		}
		return result;
	}

	/**
	 * close current shape or list of points
	 * 
	 */
	public void close() {
		int nbpt = currentpoints.size();

		if (nbpt > 0) {
			int[] polix, poliy;
			polix = new int[nbpt];
			poliy = new int[nbpt];
			Point p;
			List vset = null;
			if (Character.isUpperCase(currenttype)) {
				vset = (List) verticalSets.get(String.valueOf(currenttype));
				if (vset == null) {
					vset = new ArrayList();
					verticalSets.put(String.valueOf(currenttype), vset);
				}
			}
			for (int i = 0; i < currentpoints.size(); i++) {
				p = (Point) currentpoints.get(i);
				polix[i] = (int) p.getX();
				poliy[i] = (int) p.getY();
				if (Character.isUpperCase(currenttype) && vset != null) {
					vset.add(p);
				}
			}
			if (Character.isLowerCase(currenttype)) {
				List list = getSetInstance(currenttype);
				if (currenttype > 'f') {
					poli = new Polygon(polix, poliy, nbpt);
					list.add(poli);
				} else {
					if (currentpoints.size() >= 2) {
						Point p1, p2;
						p1 = (Point) currentpoints.get(0);
						p2 = (Point) currentpoints.get(1);
						Rectangle rect = new Rectangle();
						rect.setFrameFromDiagonal(p1, p2);
						list.add(rect);
					}
				}
			}
			currentpoints.clear();
		}
	}

	/**
	 * return list of specified type (creates it if it does not exists)
	 * 
	 * @param type
	 *            the type
	 * @return the list
	 */
	private List getSetInstance(char type) {
		List set = getSet(type);
		if (set == null) {
			if (type >= 'a' && type <= 'f') {
				sides.add(Character.toString(type).toUpperCase());
			}
			String key = String.valueOf(type);
			set = new ArrayList();
			allSets.put(key, set);
		}
		return set;
	}

	/**
	 * return list of specified type
	 * 
	 * @param type
	 *            the type
	 * @return the list
	 */
	public List getSet(char type) {
		return (List) allSets.get(String.valueOf(type));
	}

	/**
	 * Returns the currenttype.
	 * 
	 * @return char
	 */
	public char getCurrenttype() {
		return currenttype;
	}

	/**
	 * Sets the currenttype.
	 * 
	 * @param currenttype
	 *            The currenttype to set
	 */
	public void setCurrenttype(char currenttype) {
		this.currenttype = currenttype;
		currentpoints = new ArrayList();
	}

	/**
	 * get maps length in game unit
	 * 
	 * @return map length
	 */
	public int getMapLength() {
		return mapLength;
	}

	/**
	 * get map width in game unit
	 * 
	 * @return map width
	 */
	public int getMapWidth() {
		return mapWidth;
	}

	/**
	 * Set the mapLength.
	 * 
	 * @param mapLength
	 *            mapLength to set
	 */
	public void setMapLength(int mapLength) {
		this.mapLength = mapLength;
	}

	/**
	 * Set the mapWidth.
	 * 
	 * @param mapWidth
	 *            mapWidth to set
	 */
	public void setMapWidth(int mapWidth) {
		this.mapWidth = mapWidth;
	}

	/**
	 * return list of all sides available on this map
	 * 
	 * @return list of all sides
	 */
	public ArrayListModel getSides() {
		return sides;
	}

	/**
	 * is map flipped from left to right
	 * 
	 * @return true if map is flipped from left to right
	 */
	public boolean isFlipLR() {
		return flipLR;
	}

	/**
	 * is map flipped from up to down
	 * 
	 * @return true if map is flipped from up to down
	 */
	public boolean isFlipUD() {
		return flipUD;
	}

	/**
	 * name of map image
	 * 
	 * @return map image name
	 */
	public String getImageName() {
		return imageName;
	}

	/**
	 * set map image name
	 * 
	 * @param string
	 *            new map image name
	 */
	public void setImageName(String string) {
		imageName = string;
	}

	/**
	 * @return Returns the path.
	 */
	public String getPath() {
		return path;
	}

	/**
	 * @return Returns the currentpoints.
	 */
	public List getCurrentpoints() {
		return currentpoints;
	}

	/**
	 * @return Returns the allSets.
	 */
	public Hashtable getAllSets() {
		return allSets;
	}

	/**
	 * @return Returns the verticalSets.
	 */
	public Hashtable getVerticalSets() {
		return verticalSets;
	}

}