/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefield;

import java.awt.Graphics;
import java.io.File;
import java.util.Hashtable;

import fb.util.FBImage;
import fb.util.Resources;

/**
 * a sprite (set of image used to draw an animated 2D character)
 * 
 * @author farfpeuf
 * 
 */
public class Sprite {


	//image table used by sprite
	private FBImage imageTable;

	//images sequences available for this sprite
	protected Hashtable sequences = new Hashtable();

	//sprites instances (singleton)
	private static Hashtable instances = new Hashtable();
	
	/**
	 * constructor: initializes and the images
	 * 
	 * @param spritePath
	 *            path to the image table
	 * @param numberOfFrames
	 *            number of frames in the table
	 */
	public Sprite(String spritePath, int numberOfFrames) {
		imageTable = new FBImage(Resources.getImage(spritePath), numberOfFrames);
	}

	/**
	 * add a new image sequence to this sprite
	 * 
	 * @param name sequence name
	 * @param index1 first frame index
	 * @param index2 second frame index
	 * @param index3 third frame index
	 * @param index4 fourth frame index
	 */
	public void addSpritsequence(String name, int index1, int index2,
			int index3, int index4) {
		sequences.put(name, new SpriteSequence(index1, index2, index3, index4));
	}

	/**
	 * get singleton instance of a sprite for a specified character
	 * @param element the character
	 * @return the sprite instance
	 */
	public static Sprite getInstance(PaintedWithSprite element) {
		return getInstance(element, "");
	}

	
	/**
	 * get singleton instance of a sprite for a specified character
	 * @param element the character
	 * @param spriteExtension used if the character uses more than one sprite (like dragons)
	 * @return the sprite instance
	 */
	public static Sprite getInstance(PaintedWithSprite element, String spriteExtension) {

		String spritePath = "sprites" + File.separator
				+ element.getSpriteID() + spriteExtension + ".gif";
		if (instances.containsKey(spritePath)) {
			return (Sprite) instances.get(spritePath);
		}
		Sprite sprite;

		sprite = new Sprite(spritePath, element.getNumberOfFrames());

		instances.put(spritePath, sprite);
		element.addSpriteSequences(sprite);
		return sprite;
	}

	/**
	 * return width of a single frame in the image table
	 * @return width
	 */
	public int getFrameWitdh() {
		return imageTable.getGameWidth();
	}
	/**
	 * return height of  a single frame in the image table
	 * @return  height
	 */
	public int getFrameHeight() {
		return imageTable.getGameHeight();
	}

	/**
	 * draw one of the sprite image
	 * @param g graphics
	 * @param x x coordinate
	 * @param y y coordinate
	 * @param scale view scale
	 * @param dx x offset
	 * @param dy y offset
	 * @param index frame index
	 * @param entityscale character scale
	 */
	public void drawFrame(Graphics g, int x, int y, float scale, int dx,
			int dy, int index, float entityscale) {

		imageTable.drawFrame(g, x, y, scale, dx, dy, index, entityscale);

	}

	/**
	 * free memory used by sprites instances
	 *
	 */
	public static void free() {
		instances.clear();
	}


	/**
	 * return the image sequences available for this sprite
	 * @return the image sequences
	 */
	public Hashtable getSequences() {
		return sequences;
	}
}