/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.network;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import fb.gui.panels.MapChooser;
import fb.main.FB;

/**
 * server for network games
 * 
 * @author farfpeuf
 * 
 */
public class Server extends Thread {
	/**
	 * socket port number
	 */
	private int socketNumber = 10409;

	/**
	 * clients list
	 */
	private static List clientList = Collections
			.synchronizedList(new LinkedList());

	/**
	 * maximu number of clients
	 */
	private static int MAX_CLIENTS = 4;

	/**
	 * did the game started ?
	 */
	private static boolean gameStarted = false;

	/**
	 * stop the server ?
	 */
	private static boolean stop = false;

	/**
	 * socket used to send and receive network messages
	 */
	private ServerSocket serverSocket;

	/**
	 * next client id
	 */
	private int nextId = 0;

	/**
	 * singleton instance
	 */
	private static Server instance;

	/**
	 * get singleton instance
	 * 
	 * @return Returns the singleton.
	 */
	public static Server getInstance() {
		if (instance == null) {
			instance = new Server();
		}
		return instance;
	}

	/**
	 * delete singleton to free memory
	 * 
	 */
	public static synchronized void freeInstance() {
		if (null != instance) {
			instance.destroy();
			instance = null;
		}
	}

	/**
	 * @return true if instance exists
	 */
	public static synchronized boolean hasInstance() {
		return (null != instance);
	}

	/**
	 * constructor initialize thread name
	 */
	private Server() {
		super();
		setName("Server thread");
	}

	/**
	 * @param numsock
	 *            The numsock to set.
	 */
	public void setSocketNumber(int numsock) {
		this.socketNumber = numsock;
	}

	/**
	 * server thread loop
	 */
	public void run() {

		try {
			serverSocket = new ServerSocket(socketNumber, 0);
			// FB.debug("Server started on port " + numsock);

			try {

				while (!stop) {

					Socket clientSocket = serverSocket.accept();

					try {

						if (clientList.size() >= MAX_CLIENTS || gameStarted) {

							new PrintWriter(new BufferedWriter(
									new OutputStreamWriter(clientSocket
											.getOutputStream())), true)
									.println("GAMEFULL");
							throw new IOException();

						}

						ClientHandler clientHandler = new ClientHandler(
								clientSocket);
						clientList.add(clientHandler);
						clientHandler.start();
						if (!FB.getInstance().isBatch()) {
							clientHandler.write(MessagesConstants.SETMAP
									+ MapChooser.getInstance(true)
											.getSelectedMap());
						}
						clientHandler
								.write(MessagesConstants.YOURIDIS + nextId);
						broadcast(MessagesConstants.CONNECTION
								+ clientSocket.getInetAddress().getHostName()
								+ "#" + nextId);
						nextId++;

					} catch (IOException ioe) {

						clientSocket.close();

					}

				}

			} finally {

				serverSocket.close();

			}
		} catch (SocketException soe) {
			FB.exception(soe);
		} catch (IOException ioe) {
			FB.exception(ioe);
		}

	}

	/**
	 * stop client connections and start the game
	 * 
	 */
	public void stopConnections() {

		gameStarted = true;
		broadcast(MessagesConstants.INITGAME + clientList.size() + "#"
				+ System.currentTimeMillis());
	}

	/**
	 * broadcast a message
	 * 
	 * @param str
	 *            the message to broadcast
	 */
	public static void broadcast(String str) {

		// FB.debug("broadcast "+str);
		synchronized (clientList) {
			Iterator i = clientList.iterator();
			while (i.hasNext()) {

				ClientHandler next = (ClientHandler) i.next();
				next.write(str);

			}

		}
	}

	/**
	 * remove a client from client list
	 * 
	 * @param clientHandler
	 *            the client handler
	 */
	public static void removeClient(ClientHandler clientHandler) {
		synchronized (clientList) {
			clientList.remove(clientHandler);
			clientList.notify();
		}
	}

	/**
	 * destroy the server
	 */
	public void destroy() {
		stop = true;
		if (serverSocket != null) {
			try {
				serverSocket.close();
			} catch (IOException e) {
				FB.exception(e);
			}
		}

	}

}

/**
 * slient handler
 * 
 * @author farfpeuf
 * 
 */
class ClientHandler extends Thread {
	/**
	 * socket used to send and receive network messages
	 */
	private Socket clientSocket;

	/**
	 * reader used to receive messages
	 */
	private BufferedReader in;

	/**
	 * writer used to send messages
	 */
	private PrintWriter out;

	/**
	 * messages queue
	 */
	private MessagesQueue messagesQueue;

	/**
	 * stop the handler ?
	 */
	private boolean stop = false;

	// each client as its queue and its sending thread
	/**
	 * constructor
	 * 
	 * @param clientSocket
	 *            the client socket
	 * @throws IOException
	 */
	public ClientHandler(Socket clientSocket) throws IOException {

		this.clientSocket = clientSocket;

		in = new BufferedReader(new InputStreamReader(clientSocket
				.getInputStream()));

		out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(
				clientSocket.getOutputStream())), true);

		out.println();

		messagesQueue = new MessagesQueue();

	}

	/**
	 * hanlder loop
	 */
	public void run() {
		setName("client-handler receiver thread");
		new Sender(this);

		try {

			while (!stop && !isInterrupted()) {

				String str = in.readLine();
				if (str == null)
					break; // EOF
				Server.broadcast(str);

				yield();

			}

		} catch (IOException lostConnection) {
			Server.broadcast(MessagesConstants.DISCONNECT
					+ clientSocket.getInetAddress().getHostName());

		} finally {

			try {
				clientSocket.close();
			} catch (IOException ignore) {
			}

			Server.removeClient(this);

		}

	}

	/**
	 * destroy handler
	 */
	public void destroy() {
		stop = true;
		if (clientSocket != null) {
			try {
				clientSocket.close();
			} catch (IOException e) {
				FB.exception(e);
			}
		}
	}

	/**
	 * enqueue a message
	 * 
	 * @param str
	 *            the message
	 */
	public void write(String str) {

		messagesQueue.enqueue(str);

	}

	/**
	 * a client handler sender thread
	 * 
	 * @author farfpeuf
	 * 
	 */
	class Sender extends Thread {
		/**
		 * the client handler
		 */
		private Thread handler;

		/**
		 * constructor
		 * 
		 * @param clientHandler
		 *            the client handler
		 */
		public Sender(Thread clientHandler) {

			handler = clientHandler;
			start();

		}

		/**
		 * sender loop
		 */
		public void run() {
			setName("client-handler sender thread");
			while (!out.checkError()) {

				while (messagesQueue.size() > 0) {
					out.write(messagesQueue.dequeue() + "\n");
				}

				try {
					sleep((long) 10);
				} catch (InterruptedException ie) {
				}

			}

			handler.interrupt();

		}

	}

}